# --
# Copyright (C) 2001-2024 OTRS AG, https://otrs.com/
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (GPL). If you
# did not receive this file, see https://www.gnu.org/licenses/gpl-3.0.txt.
# --

package Kernel::System::CheckEnvironment;

## nofilter(TidyAll::Plugin::OTRS::Perl::Require)

use strict;
use warnings;

use parent qw(Kernel::System::CheckEnvironment::Base);

=head1 NAME

Kernel::System::CheckEnvironment - Run all environment check modules.

=head1 PUBLIC INTERFACE

=head2 new()

Don't use the ObjectManager, use it directly instead:

    use Kernel::System::CheckEnvironment;
    my $CheckEnvironmentObject = Kernel::System::CheckEnvironment->new();
    $CheckEnvironmentObject->Run();

The ObjectManager should not be used especially when using this module from a script,
because we want this to always work, no matter if any dependencies fail.

=cut

sub new {
    my ( $Type, %Param ) = @_;

    my $Self = {};
    bless( $Self, $Type );

    return $Self;
}

sub Run {
    my ( $Self, %Param ) = @_;

    if ( $Param{Options}->{Help} ) {
        return $Self->Help();
    }

    my $Error;
    my @Output;

    # Define environment checks.
    my @Checks = (
        {
            Name   => 'Perl Modules',
            Module => 'PerlModules',
        },
        {
            Name   => 'External Programs',
            Module => 'ExternalPrograms',
        },
    );

    ENV_CHECK:
    for my $Check (@Checks) {

        next ENV_CHECK if !$Check->{Name} || !$Check->{Module};

        if ( $Param{Options}->{PackageList} ) {
            push @Output, {
                State   => 'Info',
                Message => "Command(s) for installing required $Check->{Name}:",
            };
        }
        else {
            push @Output, {
                State   => 'Info',
                Message => "Checking for $Check->{Name}:",
            };
        }

        my $ModuleName = "Kernel::System::CheckEnvironment::$Check->{Module}";
        my $Module     = $ModuleName;
        $Module =~ s{::}{/}g;
        $Module .= '.pm';

        my $ModuleObject;

        eval {
            require $Module;    ## no critic
            $ModuleObject = $ModuleName->new( Options => $Param{Options} );
        };

        if ( !$ModuleObject ) {
            push @Output, {
                State   => 'Error',
                Message => "Could not instantiate '$Check->{Module}' check:\n$@",
            };
            $Error = 1;
            last ENV_CHECK;
        }

        eval {
            my $Check = $ModuleObject->Run();
            if ( !$Check->{Success} ) {
                $Error = 1;
            }

            push @Output, {
                State   => $Error ? 'Error' : 'OK',
                Message => $Check->{Message},
            };
        };

        if ($@) {
            push @Output, {
                State   => 'Error',
                Message => "Could not run '$Check->{Module}' check!",
            };
            last ENV_CHECK;
        }
    }

    if ($Error) {
        push @Output, {
            State   => 'Error',
            Message => 'Environment checks encountered some errors, please see above!',
        };
    }

    return (
        Error  => $Error,
        Output => \@Output,
    );
}

1;
