# --
# Copyright (C) 2001-2024 OTRS AG, https://otrs.com/
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (GPL). If you
# did not receive this file, see https://www.gnu.org/licenses/gpl-3.0.txt.
# --

package Kernel::System::CheckEnvironment::Base;

use strict;
use warnings;

=head1 NAME

Kernel::System::CheckEnvironment::Base - Base class for environment check script.

=head1 DESCRIPTION

Inherit the class via the parent statement:

    use parent Kernel::System::CheckEnvironment::Base;

=cut

sub _VersionClean {
    my ( $Self, %Param ) = @_;

    if ( !$Param{Version} || $Param{Version} eq 'undef' ) {
        return {
            Version  => 0,
            Operator => undef,
        };
    }

    my ( $Operator, $Version ) = $Param{Version} =~ m{ (^ [!><=]{1,2})? \s* (.+) $}xs;
    $Operator //= '=';

    # Replace all special characters with a dot.
    $Version =~ s{ [_-] }{.}xmsg;

    # Cut leading 'v', e.g. from 'v4.25.4'.
    $Version =~ s{^v}{}xms;

    my @VersionParts = split q{\.}, $Version;

    my $CleanedVersion = '';
    for my $Count ( 0 .. 4 ) {
        $VersionParts[$Count] ||= 0;
        $CleanedVersion .= sprintf "%04d", $VersionParts[$Count];
    }

    return {
        Version  => ( int $CleanedVersion ),
        Operator => $Operator,
    };
}

sub _VersionCompare {
    my ( $Self, %Param ) = @_;

    # equal
    if ( $Param{Operator} eq '==' || $Param{Operator} eq '=' ) {
        return $Param{ActualVersion} == $Param{WantedVersion};
    }

    # greater
    if ( $Param{Operator} eq '>' ) {
        return $Param{ActualVersion} > $Param{WantedVersion};
    }

    # less
    if ( $Param{Operator} eq '<' ) {
        return $Param{ActualVersion} < $Param{WantedVersion};
    }

    # greater equal
    if ( $Param{Operator} eq '>=' ) {
        return $Param{ActualVersion} >= $Param{WantedVersion};
    }

    # less equal
    if ( $Param{Operator} eq '<=' ) {
        return $Param{ActualVersion} >= $Param{WantedVersion};
    }

    # not equal
    if ( $Param{Operator} eq '!=' ) {
        return $Param{ActualVersion} >= $Param{WantedVersion};
    }

    return 0;
}

1;
