# --
# Copyright (C) 2001-2024 OTRS AG, https://otrs.com/
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (GPL). If you
# did not receive this file, see https://www.gnu.org/licenses/gpl-3.0.txt.
# --

package Kernel::System::CheckEnvironment::ExternalPrograms;

use strict;
use warnings;

use parent qw(Kernel::System::CheckEnvironment::Base);

use Kernel::System::VariableCheck qw( :all );

=head1 NAME

Kernel::System::CheckEnvironment::ExternalPrograms - Environment check for required external programs.

=head1 PUBLIC INTERFACE

=head2 new()

Don't use the ObjectManager, use it directly instead:

    use Kernel::System::CheckEnvironment::ExternalPrograms;
    my $ExternalProgramsObject = Kernel::System::CheckEnvironment::ExternalPrograms->new();
    $ExternalProgramsObject->Run();

The ObjectManager should not be used especially when using this module from a script,
because we want this to always work, no matter if any dependencies fail.

=cut

sub new {
    my ( $Type, %Param ) = @_;

    my $Self = {};
    bless( $Self, $Type );

    $Self->{Options} = $Param{Options} // {};

    $Self->{InstTypeToCMD} = {

        # [InstType] => {
        #    CMD => '[cmd to install program]',
        # }
        aptget => {
            CMD => 'apt-get install -y %s',
        },
        dnf => {
            CMD => 'dnf install %s',
        },
        emerge => {
            CMD => 'emerge %s',
        },
        yum => {
            CMD => 'yum install %s',
        },
        zypper => {
            CMD => 'zypper install %s',
        },
        ports => {
            CMD => 'ports install %s',
        },
        default => {
            CMD => '<package-manager> install %s',
        },
    };

    $Self->{DistToInstType} = {

        # apt-get
        debian => 'aptget',
        ubuntu => 'aptget',

        # emerge
        #   For reasons unknown, some environments return "gentoo" (incl. the quotes).
        '"gentoo"' => 'emerge',
        gentoo     => 'emerge',

        # yum
        centos => 'yum',
        fedora => 'yum',
        rhel   => 'yum',
        redhat => 'dnf',

        # zypper
        suse => 'zypper',

        # FreeBSD
        freebsd => 'ports',
    };

    eval {
        require Linux::Distribution;    ## nofilter(TidyAll::Plugin::OTRS::Perl::Require)
        import Linux::Distribution;
        $Self->{OSDist} = Linux::Distribution::distribution_name() || '';
    };
    if ( !defined $Self->{OSDist} ) {
        $Self->{OSDist} = $^O;
    }

    return $Self;
}

sub Run {
    my ( $Self, %Param ) = @_;

    my @NeededPrograms = (
        {
            Name         => 'GnuPG',
            Bin          => 'gpg',
            Required     => 0,
            Comment      => 'Required for PGP support.',
            VersionParam => '--version',
            VersionRegex => qr{ (?'Variant' (?<=/) \w+ )? \)? \s* (?'Version' ( [0-9]+ \.? )+ ) $ }xms,
            InstTypes    => {
                aptget => 'gnupg',
                dnf    => 'gpg',
                emerge => 'app-crypt/gnupg',
                yum    => 'gpg',
                zypper => 'gpg',
                ports  => 'gnupg',
            },
        },
        {
            Name         => 'npm',
            Bin          => 'npm',
            Required     => 1,
            Comment      => 'Required for building application frontends. Check https://nodejs.org/ on how to install.',
            Version      => '>= 6.14.8',
            VersionParam => '-v',
            VersionRegex => qr{ (?'Version' ( [0-9]+ \.? )+ ) }xms,
            InstTypes    => {
                aptget => 'npm',
                dnf    => 'npm',
                emerge => 'net-libs/nodejs',
                ppm    => 'npm',
                yum    => 'npm',
                zypper => 'npm',
                ports  => 'npm',
            },
            Depends => [
                {
                    Name     => 'Node.js',
                    Bin      => 'node',
                    AltBins  => ['nodejs'],
                    Required => 1,
                    Comment  => 'JavaScript run-time environment. Check https://nodejs.org/ on how to install.',

                    # Test.
                    Version         => [ '>= 20.4.0', '< 21.0.0' ],
                    VersionOperator => 'AND',
                    VersionParam    => '-v',
                    VersionRegex    => qr{ (?'Version' ( [0-9]+ \.? )+ ) }xms,
                    InstTypes       => {
                        aptget => 'nodejs',
                        dnf    => 'nodejs',
                        emerge => 'net-libs/nodejs',
                        ppm    => 'nodejs',
                        yum    => 'nodejs',
                        zypper => 'nodejs',
                        ports  => 'nodejs',
                    },
                },
                {
                    Name         => 'Python',
                    Bin          => 'python3',
                    AltBins      => ['python'],
                    Required     => 1,
                    Comment      => 'Required to build some npm dependencies. Check https://www.python.org/downloads.',
                    Version      => '>= 3.7.1',
                    VersionParam => '--version',
                    VersionRegex => qr{ (?'Version' ( [0-9]+ \.? )+ ) }xms,
                    InstTypes    => {
                        aptget => 'python3',
                        dnf    => 'python3',
                        emerge => 'dev-lang/python',
                        ppm    => 'python3',
                        yum    => 'python3',
                        zypper => 'python3',
                        ports  => 'python3',
                    },
                },
                {
                    Name         => 'g++',
                    Bin          => 'g++',
                    AltBins      => ['gpp'],
                    Required     => 1,
                    Comment      => 'Required to build npm dependencies.',
                    Version      => '>= 8.5.0',
                    VersionParam => '--version',
                    VersionRegex => qr{ (?'Version' ( [0-9]+ \.? )+ ) }xms,
                    InstTypes    => {
                        aptget => 'g++',
                        dnf    => 'gcc-c++',
                        emerge => 'sys-devel/gcc',
                        ppm    => 'g++',
                        yum    => 'gcc-c++',
                        zypper => 'gcc-c++',
                        ports  => 'g++',
                    },
                },
            ],
        },
        {
            Name     => 'OpenSSL',
            Bin      => 'openssl',
            Required => 0,
            Comment  => 'Required for S/MIME support.',
            Version  => {

                # FIXME: If Debian 9 is (really) out of support, set openSSL version to >= 1.1.1.
                'OpenSSL'  => [ '1.0.2', '>= 1.1.0' ],
                'LibreSSL' => ['<= 2.8.3'],
            },
            VersionRecommended => {
                'OpenSSL' => [
                    {
                        Version => '>= 1.1.1',
                        Comment =>
                            'All OpenSSL versions before 1.1.1 are out of support and no longer receiving updates. Extended support is available for 1.0.2 from OpenSSL Software Services for premium support customers.',
                    },
                ],
            },
            VersionParam => 'version -v',
            VersionRegex => qr{ (?'Variant' \w+ )? \s* (?'Version' ( [0-9]+ \.? )+ ) }xms,
            InstTypes    => {
                aptget => 'openssl',
                dnf    => 'openssl',
                emerge => 'dev-libs/openssl',
                yum    => 'openssl',
                zypper => 'openssl',
                ports  => 'openssl',
            },
        },
    );

    my $Message;
    my $Error;

    if ( $Self->{Options}->{PackageList} ) {
        my %PackageList = $Self->_PackageList( \@NeededPrograms );

        if ( IsArrayRefWithData( $PackageList{Packages} ) ) {

            my $CMD = $PackageList{CMD};

            for my $Package ( @{ $PackageList{Packages} } ) {
                if ( $PackageList{SubCMD} ) {
                    $Package = sprintf $PackageList{SubCMD}, $Package;
                }
            }
            my $CMDText = sprintf $CMD, join( ' ', @{ $PackageList{Packages} } );
            $Message .= "$CMDText\n";
        }
    }
    else {
        my $Depends = 0;

        for my $Program (@NeededPrograms) {
            my $Check = $Self->_CheckProgram( $Program, $Depends );
            if ( !$Check->{Success} ) {
                $Error = 1;
            }
            $Message .= $Check->{Message};
        }
    }

    return {
        Success => $Error ? 0 : 1,
        Message => $Message,
    };
}

sub _CheckProgramVersion {
    my ( $Self, $Program, $ProgramBinary ) = @_;

    # All program versions are valid.
    if ( !$Program->{Version} ) {
        return {
            Success       => 1,
            StatusMessage => "<green>ok</green>\n",
        };
    }

    my $Comment  = $Program->{Comment}  ? ' - ' . $Program->{Comment} : '';
    my $Required = $Program->{Required} ? 'required'                  : 'optional';

    # Check the version.
    my $BinaryVersion = `$ProgramBinary $Program->{VersionParam}`;
    chomp $BinaryVersion;

    # If version can not be determined, output some info and exit.
    if ( !$BinaryVersion ) {
        return {
            Success       => 0,
            StatusMessage => "<red>Binary found, but version check was unsuccessful! ($Required$Comment)</red>\n"
        };
    }

    my $VersionText = $BinaryVersion;
    my $BinaryVariant;

    if ( $Program->{VersionRegex} && $BinaryVersion =~ $Program->{VersionRegex} ) {
        $VersionText = $BinaryVersion = $+{Version};
        if ( $+{Variant} ) {
            $VersionText   = $BinaryVersion . '/' . $+{Variant};
            $BinaryVariant = $+{Variant};
        }
    }

    my $StatusMessage = "<green>ok</green> (v$VersionText)\n";

    # Cleanup binary version number.
    my $CleanedBinaryVersion = $Self->_VersionClean(
        Version => $BinaryVersion,
    )->{Version};

    my $ProgramVersionObject;
    if ( IsHashRefWithData( $Program->{Version} ) ) {
        $ProgramVersionObject = $Program->{Version}->{ $+{Variant} };
    }
    else {
        $ProgramVersionObject = $Program->{Version};
    }

    my @ProgramVersions;
    if ( IsArrayRefWithData($ProgramVersionObject) ) {
        @ProgramVersions = @{$ProgramVersionObject};
    }
    else {
        @ProgramVersions = ($ProgramVersionObject);
    }

    my $VersionOperator = ( defined $Program->{VersionOperator} ) ? lc $Program->{VersionOperator} : 'or';

    my ( $ProgramVersion, $Operator );
    VERSION:
    for my $Version (@ProgramVersions) {
        my $CleanedVersion = $Self->_VersionClean(
            Version => $Version,
        );

        $Operator = $CleanedVersion->{Operator};
        my $CleanedProgramVersion = $CleanedVersion->{Version};

        my $VersionFound = $Self->_VersionCompare(
            Operator      => $Operator,
            ActualVersion => $CleanedBinaryVersion,
            WantedVersion => $CleanedProgramVersion,
        );

        if ( $VersionOperator eq 'or' ) {
            if ($VersionFound) {
                $ProgramVersion = $Version;
                last VERSION;
            }
        }
        elsif ( $VersionOperator eq 'and' ) {
            if ($VersionFound) {
                $ProgramVersion = $Version;
            }
            else {
                $ProgramVersion = undef;
                last VERSION;
            }
        }
    }

    if ( !$ProgramVersion ) {
        return {
            StatusMessage => (
                sprintf "<red>Version %s installed but %s is required! (%s%s)</red>\n",
                $BinaryVersion, ( join " $VersionOperator ", @ProgramVersions ), $Required, $Comment,
            ),
            Success => 0
        };
    }

    my $VersionRecommendedComment = '';
    if ( $Program->{VersionRecommended} ) {
        my @VersionRecommendedObject;
        if ($BinaryVariant) {
            @VersionRecommendedObject = @{ $Program->{VersionRecommended}->{$BinaryVariant} };
        }
        else {
            @VersionRecommendedObject = @{ $Program->{VersionRecommended} };
        }

        VERSIONRECOMMENDED:
        for my $VersionRecommended (@VersionRecommendedObject) {
            my $CleanedVersionRecommended = $Self->_VersionClean(
                Version => $VersionRecommended->{Version},
            );

            my $VersionIsRecommended = $Self->_VersionCompare(
                Operator      => $CleanedVersionRecommended->{Operator},
                ActualVersion => $CleanedBinaryVersion,
                WantedVersion => $CleanedVersionRecommended->{Version},
            );

            if ( !$VersionIsRecommended ) {
                $VersionRecommendedComment = sprintf "<yellow>%s</yellow>", $VersionRecommended->{Comment};
                last VERSIONRECOMMENDED;
            }
        }

    }

    return {
        StatusMessage => "<green>ok</green> ($VersionText)\n" . $VersionRecommendedComment,
        Success       => 1
    };
}

sub _CheckProgram {
    my ( $Self, $Program, $Depends ) = @_;

    my $Message = "  " x ( $Depends + 1 );
    $Message .= "o $Program->{Name}";
    my $Length = 33 - ( length( $Program->{Name} ) + ( $Depends * 2 ) );
    $Message .= '.' x $Length;

    my $Error;

    # Check if binary or alternative exists.
    my $ProgramBinary;
    BINARY:
    for my $Binary ( $Program->{Bin}, @{ $Program->{AltBins} } ) {
        $ProgramBinary = `which $Binary`;
        chomp $ProgramBinary;

        last BINARY if $ProgramBinary;
    }

    my $Comment  = $Program->{Comment}  ? ' - ' . $Program->{Comment} : '';
    my $Required = $Program->{Required} ? 'required'                  : 'optional';

    # If binary is not found, output some info about it and exit.
    if ( !$ProgramBinary ) {
        my $Color = 'yellow';

        # OS Install Command
        my %InstallCommand = $Self->_GetInstallCommand($Program);

        # Create example installation string for the program.
        my $InstallText = '';
        if ( IsHashRefWithData( \%InstallCommand ) ) {
            my $CMD = $InstallCommand{CMD};
            if ( $InstallCommand{SubCMD} ) {
                $CMD = sprintf $InstallCommand{CMD}, $InstallCommand{SubCMD};
            }

            $InstallText = " Use: '" . sprintf( $CMD, $InstallCommand{Package} ) . "'";
        }

        if ( $Program->{Required} ) {
            $Color = 'red';
            $Error = 1;
        }

        $Message .= "<$Color>Not installed!</$Color>$InstallText ($Required$Comment)\n";

        return {
            Success => $Error ? 0 : 1,
            Message => $Message,
        };
    }

    my $Result = $Self->_CheckProgramVersion( $Program, $ProgramBinary );
    $Message .= $Result->{StatusMessage};

    if ( !$Result->{Success} ) {
        return {
            Success => 0,
            Message => $Message,
        };
    }

    if ( $Program->{Depends} ) {
        for my $ProgramSub ( @{ $Program->{Depends} } ) {
            my $DependencyCheck = $Self->_CheckProgram( $ProgramSub, $Depends + 1 );
            if ( !$DependencyCheck->{Success} ) {
                $Error = 1;
            }
            $Message .= $DependencyCheck->{Message};
        }
    }

    return {
        Success => $Error ? 0 : 1,
        Message => $Message,
    };
}

sub _PackageList {
    my ( $Self, $PackageList ) = @_;

    my $CMD;
    my $SubCMD;
    my @Packages;

    PROGRAM:
    for my $Program ( @{$PackageList} ) {

        my $Required = $Program->{Required};

        # Check if binary exists.
        my $ProgramBinary = `which $Program->{Bin}`;
        chomp $ProgramBinary;

        # Binary may live with another name, make sure to check all the alternatives.
        if ( !$ProgramBinary ) {
            ALTBIN:
            for my $AltBin ( @{ $Program->{AltBins} } ) {
                $ProgramBinary = `which $AltBin`;
                chomp $ProgramBinary;

                last ALTBIN if $ProgramBinary;
            }
        }

        if ( !$ProgramBinary ) {
            my %InstallCommand = $Self->_GetInstallCommand($Program);

            if ( $Program->{Depends} ) {

                PROGRAMSUB:
                for my $ProgramSub ( @{ $Program->{Depends} } ) {
                    my $Required          = $Program->{Required};
                    my %InstallCommandSub = $Self->_GetInstallCommand($ProgramSub);

                    next PROGRAMSUB if !IsHashRefWithData( \%InstallCommandSub );
                    next PROGRAMSUB if !$Required;

                    push @Packages, $InstallCommandSub{Package};
                }
            }

            next PROGRAM if !IsHashRefWithData( \%InstallCommand );
            next PROGRAM if !$Required;

            $CMD    = $InstallCommand{CMD};
            $SubCMD = $InstallCommand{SubCMD};
            push @Packages, $InstallCommand{Package};
        }
    }

    return (
        CMD      => $CMD,
        SubCMD   => $SubCMD,
        Packages => \@Packages,
    );
}

sub _GetInstallCommand {
    my ( $Self, $Program ) = @_;

    my $CMD;
    my $SubCMD;
    my $Package;

    # Returns the installation type e.g. ppm.
    my $InstType      = $Self->{DistToInstType}->{ $Self->{OSDist} };
    my $OutputInstall = 1;

    if ($InstType) {

        # Gets the install command for installation type:
        #   e.g. apt-get install -y %s
        # Default is to just list the program names:
        #   e.g. %s
        $CMD    = $Self->{InstTypeToCMD}->{$InstType}->{CMD};
        $SubCMD = $Self->{InstTypeToCMD}->{$InstType}->{SubCMD};

        # Gets the target package.
        if (
            exists $Program->{InstTypes}->{$InstType}
            && !defined $Program->{InstTypes}->{$InstType}
            )
        {
            # if we a hash key for the installation type but a undefined value
            # then we prevent the output for the installation command
            $OutputInstall = 0;
        }
        else {
            # if the package name is defined for the installation type
            #   e.g. aptget then we use this as package name
            $Package = $Program->{InstTypes}->{$InstType};
        }
    }

    return if !$OutputInstall;

    if ( !$CMD || !$Package ) {
        $CMD     = $Self->{InstTypeToCMD}->{default}->{CMD};
        $SubCMD  = $Self->{InstTypeToCMD}->{default}->{SubCMD};
        $Package = $Program->{Bin};
    }

    return (
        CMD     => $CMD,
        SubCMD  => $SubCMD,
        Package => $Package,
    );
}

1;
