# --
# Copyright (C) 2001-2024 OTRS AG, https://otrs.com/
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (GPL). If you
# did not receive this file, see https://www.gnu.org/licenses/gpl-3.0.txt.
# --

package Kernel::System::CheckEnvironment::PerlModules;

use strict;
use warnings;

use parent qw(Kernel::System::CheckEnvironment::Base);

use Kernel::System::Environment;
use Kernel::System::VariableCheck qw( :all );

=head1 NAME

Kernel::System::CheckEnvironment::PerlModules - Environment check for required Perl modules.

=head1 PUBLIC INTERFACE

=head2 new()

Don't use the ObjectManager, use it directly instead:

    use Kernel::System::CheckEnvironment::PerlModules;
    my $PerlModulesObject = Kernel::System::CheckEnvironment::PerlModules->new();
    $PerlModulesObject->Run();

The ObjectManager should not be used especially when using this module from a script,
because we want this to always work, no matter if any dependencies fail.

=cut

sub new {
    my ( $Type, %Param ) = @_;

    my $Self = {};
    bless( $Self, $Type );

    $Self->{Options} = $Param{Options} // {};

    $Self->{InstTypeToCMD} = {

        # [InstType] => {
        #    CMD       => '[cmd to install module]',
        #    UseModule => 1/0,
        # }
        # Set UseModule to 1 if you want to use the CPAN module name of the package as replace string:
        #   e.g. yum install "perl(Date::Format)"
        # If you set it 0 it will use the name for the InstType of the module:
        #   e.g. apt-get install -y libtimedate-perl
        # and as fallback the default cpan install command:
        #   e.g. cpan DBD::Oracle
        aptget => {
            CMD       => 'apt-get install -y %s',
            UseModule => 0,
        },
        emerge => {
            CMD       => 'emerge %s',
            UseModule => 0,
        },
        yum => {
            CMD       => 'yum install "%s"',
            SubCMD    => 'perl(%s)',
            UseModule => 1,
        },
        default => {
            CMD => 'cpan %s',
        },
    };

    $Self->{DistToInstType} = {

        # apt-get
        debian => 'aptget',
        ubuntu => 'aptget',

        # emerge
        #   For reasons unknown, some environments return "gentoo" (incl. the quotes).
        '"gentoo"' => 'emerge',
        gentoo     => 'emerge',

        # yum
        centos => 'yum',
        fedora => 'yum',
        rhel   => 'yum',
        redhat => 'yum',
    };

    eval {
        require Linux::Distribution;    ## nofilter(TidyAll::Plugin::OTRS::Perl::Require)
        import Linux::Distribution;
        $Self->{OSDist} = Linux::Distribution::distribution_name() || '';
    };
    if ( !defined $Self->{OSDist} ) {
        $Self->{OSDist} = $^O;
    }

    return $Self;
}

sub Run {
    my ( $Self, %Param ) = @_;

    my @NeededModules = (
        {
            Module    => 'Archive::Zip',
            Required  => 1,
            Comment   => 'Required for compressed file generation.',
            InstTypes => {
                aptget => 'libarchive-zip-perl',
                emerge => 'dev-perl/Archive-Zip',
            },
        },
        {
            Module    => 'BSD::Resource',
            Required  => 1,
            Comment   => 'Required for Mojolicious::Plugin::Status.',
            InstTypes => {
                aptget => 'libbsd-resource-perl',
                emerge => 'dev-perl/BSD-Resource',
            },
        },
        {
            Module    => 'Compress::LZ4',
            Required  => 0,
            Comment   => 'Required for faster (de-)compressing.',
            InstTypes => {
                aptget => 'libcompress-lz4-perl',
            },
        },
        {
            Module          => 'CryptX',
            Required        => 1,
            VersionRequired => '0.067',    # required by Crypt::JWT
            InstTypes       => {
                aptget => 'libcryptx-perl',
                emerge => 'dev-perl/CryptX',
            },
        },
        {
            Module    => 'Date::Format',
            Required  => 1,
            InstTypes => {
                aptget => 'libtimedate-perl',
                emerge => 'dev-perl/TimeDate',
            },
        },
        {
            Module    => 'DateTime',
            Required  => 1,
            InstTypes => {
                aptget => 'libdatetime-perl',
                emerge => 'dev-perl/DateTime',
            },
            Depends => [
                {
                    Module              => 'DateTime::TimeZone',
                    Comment             => 'Olson time zone database, required for correct time calculations.',
                    VersionsRecommended => [
                        {
                            Version => '2.20',
                            Comment => 'This version includes recent time zone changes for Chile.',
                        },
                    ],
                    InstTypes => {
                        aptget => 'libdatetime-timezone-perl',
                        emerge => 'dev-perl/DateTime-TimeZone',
                    },
                },
            ],
        },
        {
            Module    => 'DBI',
            Required  => 1,
            InstTypes => {
                aptget => 'libdbi-perl',
                emerge => 'dev-perl/DBI',
            },
        },
        {
            Module    => 'DBD::mysql',
            Required  => 0,
            Comment   => 'Required to connect to a MySQL database.',
            InstTypes => {
                aptget => 'libdbd-mysql-perl',
                emerge => 'dev-perl/DBD-mysql',
            },
        },
        {
            Module               => 'DBD::ODBC',
            Required             => 0,
            VersionsNotSupported => [
                {
                    Version => '1.23',
                    Comment =>
                        'This version is broken and not useable! Please upgrade to a higher version.',
                },
            ],
            Comment   => 'Required to connect to a MS-SQL database.',
            InstTypes => {
                aptget => 'libdbd-odbc-perl',
            },
        },
        {
            Module    => 'DBD::Oracle',
            Required  => 0,
            Comment   => 'Required to connect to a Oracle database.',
            InstTypes => {},
        },
        {
            Module          => 'DBD::Pg',
            VersionRequired => '3.5.0',
            Required        => 0,
            Comment         => 'Required to connect to a PostgreSQL database.',
            InstTypes       => {
                aptget => 'libdbd-pg-perl',
                emerge => 'dev-perl/DBD-Pg',
            },
        },
        {
            Module    => 'Devel::NYTProf',
            Required  => 0,
            Comment   => 'Required for web server profiling.',
            InstTypes => {
                aptget => 'libdevel-nytprof-perl',
                emerge => 'dev-perl/Devel-NYTProf',
            },
        },
        {
            Module          => 'Digest::Bcrypt',
            VersionRequired => '1.209',
            Required        => 1,
            Comment         => 'For strong password hashing.',
            InstTypes       => {
                aptget => 'libdigest-bcrypt-perl',
                emerge => 'dev-perl/Digest-Bcrypt',
            },
        },
        {
            Module    => 'Email::Address::XS',
            Required  => 1,
            Comment   => 'Required to handle mails recipients parsing.',
            InstTypes => {
                aptget => 'libemail-address-xs-perl',
                emerge => 'dev-perl/Email-Address-XS',
            },
        },
        {
            Module          => 'Encode::HanExtra',
            VersionRequired => '0.23',
            Required        => 0,
            Comment         => 'Required to handle mails with several Chinese character sets.',
            InstTypes       => {
                aptget => 'libencode-hanextra-perl',
                emerge => 'dev-perl/Encode-HanExtra',
            },
        },
        {
            Module    => 'EV',
            Required  => 1,
            Comment   => 'Perl interface to libev, the high performance event loop.',
            InstTypes => {
                aptget => 'libev-perl',
                emerge => 'dev-perl/EV',
            },
        },
        {
            Module    => 'File::Map',
            Required  => 1,
            Comment   => 'Required for web server profiling and Mojolicious::Plugin::Status.',
            InstTypes => {
                aptget => 'libfile-map-perl',
                emerge => 'dev-perl/File-Map',
            },
        },
        {
            Module              => 'IO::Socket::SSL',
            Required            => 1,
            VersionRequired     => '2.009',
            Comment             => 'Required for SSL connections to web and mail servers.',
            VersionsRecommended => [
                {
                    Version => '2.066',
                    Comment => 'This version fixes email sending (bug#14357).',
                },
            ],
            InstTypes => {
                aptget => 'libio-socket-ssl-perl',
                emerge => 'dev-perl/IO-Socket-SSL',
            },
        },
        {
            Module    => 'JSON::XS',
            Required  => 0,
            Comment   => 'Recommended for faster AJAX/JavaScript handling.',
            InstTypes => {
                aptget => 'libjson-xs-perl',
                emerge => 'dev-perl/JSON-XS',
            },
        },
        {
            Module    => 'List::MoreUtils',
            Required  => 1,
            Comment   => 'Required for correct display of CommunicationStream widget.',
            InstTypes => {
                aptget => 'liblist-moreutils-perl',
                emerge => 'dev-perl/List-MoreUtils',
            },
        },
        {
            Module    => 'LWP::UserAgent',
            Required  => 1,
            InstTypes => {
                aptget => 'libwww-perl',
                emerge => 'dev-perl/libwww-perl',
            },
        },
        {
            Module          => 'Mail::IMAPClient',
            VersionRequired => '3.22',
            Comment         => 'Required for IMAP TLS connections.',
            Required        => 0,
            InstTypes       => {
                aptget => 'libmail-imapclient-perl',
                emerge => 'dev-perl/Mail-IMAPClient',
            },
            Depends => [
                {
                    Module    => 'Authen::SASL',
                    Required  => 0,
                    Comment   => 'Required for MD5 authentication mechanisms in IMAP connections.',
                    InstTypes => {
                        aptget => 'libauthen-sasl-perl',
                        emerge => 'dev-perl/Authen-SASL',
                    },
                },
                {
                    Module    => 'Authen::NTLM',
                    Required  => 0,
                    Comment   => 'Required for NTLM authentication mechanism in IMAP connections.',
                    InstTypes => {
                        aptget => 'libauthen-ntlm-perl',
                        emerge => 'dev-perl/Authen-NTLM',
                    },
                },
            ],
        },
        {
            Module    => 'Moo',
            Required  => 1,
            Comment   => 'Required for random number generator.',
            InstTypes => {
                aptget => 'libmoo-perl',
                emerge => 'dev-perl/Moo',
            },
        },
        {
            Module   => 'Moose',
            Required => 1,

 # Workaround for http://blogs.perl.org/users/ovid/2014/02/moose-dying-with-invalid-version-format-version-required.html
            VersionRequired => '2.1203',
            Comment         => 'A postmodern object system for Perl 5.',
            InstTypes       => {
                aptget => 'libmoose-perl',
                emerge => 'dev-perl/Moose',
            },
        },
        {
            Module          => 'MooseX::NonMoose',
            Required        => 0,
            VersionRequired => '0.26',
            Comment         => 'Easy subclassing of non-Moose classes.',
            InstTypes       => {
                aptget => 'libmoosex-nonmoose-perl',
                emerge => 'dev-perl/MooseX-NonMoose',
            },
        },
        {
            Module               => 'Net::DNS',
            Required             => 1,
            Comment              => 'Perl Interface to the Domain Name System.',
            VersionsNotSupported => [
                {
                    Version => '0.60',
                    Comment =>
                        'This version is broken and not useable! Please upgrade to a higher version.',
                },
            ],
            InstTypes => {
                aptget => 'libnet-dns-perl',
                emerge => 'dev-perl/Net-DNS',
            },
        },
        {
            Module    => 'Net::LDAP',
            Required  => 0,
            Comment   => 'Required for directory authentication.',
            InstTypes => {
                aptget => 'libnet-ldap-perl',
                emerge => 'dev-perl/perl-ldap',
            },
        },
        {
            Module              => 'Net::SMTP',
            Required            => 0,
            Comment             => 'Simple Mail Transfer Protocol Client.',
            VersionsRecommended => [
                {
                    # fixed version was released with perl 5.28 - can be removed if perl minimum for OTRS is >= 5.28
                    Version => '3.11',
                    Comment => 'This version fixes email sending (bug#14357).',
                },
            ],
            InstTypes => {},
        },
        {
            Module    => 'Paws',
            Required  => 0,
            Comment   => 'A Perl SDK for AWS (Amazon Web Services) APIs.',
            InstTypes => {},
        },
        {
            Module    => 'Cache::Memcached::Fast',
            Required  => 0,
            Comment   => 'Required when using memcached for caching.',
            InstTypes => {
                aptget => 'libcache-memcached-fast-perl',
            },
        },
        {
            Module    => 'Redis::Fast',
            Required  => 0,
            Comment   => 'Required for new Caching Backend',
            InstTypes => {
                aptget => 'libredis-fast-perl',
            },
        },
        {
            Module          => 'Search::Elasticsearch',
            Required        => 1,
            VersionRequired => '8.00',
            Comment         => 'Required for document search.',
            InstTypes       => {
                aptget => 'libsearch-elasticsearch-perl',
            },
        },
        {
            Module          => 'Search::Elasticsearch::Client::8_0',
            Required        => 1,
            VersionRequired => '8.00',
            Comment         => 'Required for document search.',
            InstTypes       => {
                aptget => 'libsearch-elasticsearch-perl',
            },
        },
        {
            Module    => 'Sereal',
            Required  => 0,
            Comment   => 'Required for usage of Redis cache module and for web server profiling.',
            InstTypes => {
                aptget => 'libsereal-perl',
                emerge => 'dev-perl/Sereal',
            },
        },
        {
            Module          => 'Specio',
            Required        => 1,
            VersionRequired => '0.47',     # https://github.com/houseabsolute/Specio/issues/18
            Comment   => 'Type constraints and coercions for Perl.',
            InstTypes => {
                aptget => 'libspecio-perl',
                emerge => 'dev-perl/Specio',
            },
        },
        {
            Module          => 'SQL::Abstract::More',
            Required        => 1,
            VersionRequired => '1.36',
            Comment         => 'Required to generates SQL from Perl data structures.',
            InstTypes       => {
                aptget => 'libsql-abstract-more-perl',
            },
        },
        {
            Module              => 'Template',
            Required            => 1,
            Comment             => 'Template::Toolkit, the rendering engine of OTRS.',
            VersionsRecommended => [
                {
                    Version => '2.28',
                    Comment => 'This version fixes uri and url filters.',
                },
            ],
            InstTypes => {
                aptget => 'libtemplate-perl',
                emerge => 'dev-perl/Template-Toolkit',
            },
        },
        {
            Module    => 'Template::Stash::XS',
            Required  => 1,
            Comment   => 'The fast data stash for Template::Toolkit.',
            InstTypes => {
                aptget => 'libtemplate-perl',
                emerge => 'dev-perl/Template-Toolkit',
            },
        },
        {
            Module    => 'Text::CSV_XS',
            Required  => 0,
            Comment   => 'Recommended for faster CSV handling.',
            InstTypes => {
                aptget => 'libtext-csv-xs-perl',
                emerge => 'dev-perl/Text-CSV_XS',
            },
        },
        {
            Module    => 'Text::Markdown',
            Required  => 1,
            Comment   => 'Convert Markdown syntax to (X)HTML.',
            InstTypes => {
                aptget => 'libtext-markdown-perl',
                emerge => 'dev-perl/Text-Markdown',
            },
        },
        {
            Module    => 'Time::Moment',
            Required  => 1,
            Comment   => 'Required for faster datetime calculations.',
            InstTypes => {
                aptget => 'libtime-moment-perl',
                emerge => 'dev-perl/Time-Moment',
            },
        },
        {
            Module    => 'XML::LibXML',
            Required  => 1,
            Comment   => 'Required for XML processing.',
            InstTypes => {
                aptget => 'libxml-libxml-perl',
                emerge => 'dev-perl/XML-LibXML',
            },
        },
        {
            Module    => 'XML::LibXSLT',
            Required  => 0,
            Comment   => 'Required for Generic Interface XSLT mapping module.',
            InstTypes => {
                aptget => 'libxml-libxslt-perl',
                emerge => 'dev-perl/XML-LibXSLT',
            },
        },
        {
            Module    => 'XML::Parser',
            Required  => 0,
            Comment   => 'Recommended for XML processing.',
            InstTypes => {
                aptget => 'libxml-parser-perl',
                emerge => 'dev-perl/XML-Parser',
            },
        },
        {
            Module   => 'YAML::XS',
            Required => 1,
            Comment  => 'Required for fast YAML processing.',

            # Example of how to use VersionsRecommended option.
            # VersionsRecommended => [
            #     {
            #         Version => '5.0.1',
            #         Comment => 'This version fixes a bug.',
            #     },
            #     {
            #         Version => '6.0.1',
            #         Comment => 'This version fixes a critical issue.',
            #     },
            # ],
            InstTypes => {
                aptget => 'libyaml-libyaml-perl',
                emerge => 'dev-perl/YAML-LibYAML',
            },
        },
    );

    my $Message;
    my $Error;

    if ( $Self->{Options}->{PackageList} ) {

        my %PackagesByCommand = $Self->_PackageList( \@NeededModules );

        if ( IsHashRefWithData( \%PackagesByCommand ) ) {

            for my $CMD ( sort keys %PackagesByCommand ) {

                # Wrap the package name in the the sub command
                if ( $PackagesByCommand{$CMD}->{SubCMD} ) {
                    for my $Package ( @{ $PackagesByCommand{$CMD}->{Packages} } ) {
                        $Package = sprintf $PackagesByCommand{$CMD}->{SubCMD}, $Package;
                    }
                }

                my $CMDText = sprintf $CMD, join ' ', @{ $PackagesByCommand{$CMD}->{Packages} };

                $Message .= "$CMDText\n\n";
            }
        }
    }
    else {
        my $Depends = 0;

        # Try to determine module version number.
        for my $Module (@NeededModules) {
            my $Check = $Self->_Check( $Module, $Depends );
            if ( !$Check->{Success} ) {
                $Error = 1;
            }
            $Message .= $Check->{Message};
        }

        if ( $Self->{Options}->{AllModules} ) {
            $Message .= "\nBundled modules:\n\n";

            my %PerlInfo = Kernel::System::Environment->PerlInfoGet(
                BundledModules => 1,
            );

            for my $Module ( sort keys %{ $PerlInfo{Modules} } ) {
                my $Check = $Self->_Check(
                    {
                        Module   => $Module,
                        Required => 1,
                    },
                    $Depends,
                );
                if ( !$Check->{Success} ) {
                    $Error = 1;
                }
                $Message .= $Check->{Message};
            }
        }
    }

    return {
        Success => $Error ? 0 : 1,
        Message => $Message,
    };
}

sub _Check {
    my ( $Self, $Module, $Depends ) = @_;

    # Prevent auto start of Devel::NYTProf.
    local $ENV{NYTPROF} = 'start=no' if $Module->{Module} eq 'Devel::NYTProf';

    my $Indent  = "  " x ( $Depends + 1 );
    my $Message = $Indent . "o $Module->{Module}";
    my $Length  = 33 - ( length( $Module->{Module} ) + ( $Depends * 2 ) );
    if ( $Length > 0 ) {
        $Message .= '.' x $Length;
    }

    my $Error;

    my $Version = Kernel::System::Environment->ModuleVersionGet( Module => $Module->{Module} );
    if ($Version) {

        # cleanup version number
        my $CleanedVersion = $Self->_VersionClean(
            Version => $Version,
        )->{Version};

        my $ErrorMessage;

        # Test if all module dependencies are installed by requiring the module.
        #   Don't do this for Net::DNS as it seems to take very long (>20s) in a
        #   mod_perl environment sometimes.
        my %DontRequire = (
            'Net::DNS'     => 1,
            'Email::Valid' => 1,    # uses Net::DNS internally
        );

        ## no critic
        if ( !$DontRequire{ $Module->{Module} } && !eval "require $Module->{Module}" ) {
            $ErrorMessage .= 'Not all prerequisites for this module correctly installed. ';
        }
        ## use critic

        if ( $Module->{VersionsNotSupported} ) {

            my $VersionsNotSupported = 0;
            ITEM:
            for my $Item ( @{ $Module->{VersionsNotSupported} } ) {

                my $ItemVersion = $Self->_VersionClean(
                    Version => $Item->{Version},
                )->{Version};

                if ( $CleanedVersion == $ItemVersion ) {
                    $VersionsNotSupported = $Item->{Comment};
                    last ITEM;
                }
            }

            if ($VersionsNotSupported) {
                $ErrorMessage .= "Version $Version not supported! $VersionsNotSupported ";
            }
        }

        my $AdditionalText = '';

        if ( $Module->{VersionsRecommended} ) {

            my $VersionsRecommended = 0;
            ITEM:
            for my $Item ( @{ $Module->{VersionsRecommended} } ) {

                my $ItemVersion = $Self->_VersionClean(
                    Version => $Item->{Version},
                )->{Version};

                if ( $CleanedVersion < $ItemVersion ) {
                    $AdditionalText
                        .= "$Indent  <yellow>Please consider updating to version $Item->{Version} or higher: $Item->{Comment}</yellow>\n";
                }
            }
        }

        if ( $Module->{VersionRequired} ) {

            my $RequiredModuleVersion = $Self->_VersionClean(
                Version => $Module->{VersionRequired},
            )->{Version};

            if ( $CleanedVersion < $RequiredModuleVersion ) {
                $ErrorMessage
                    .= "Version $Version installed but $Module->{VersionRequired} or higher is required! ";
            }
        }

        if ($ErrorMessage) {
            $Message .= "<red>$ErrorMessage</red>\n";
            $Error = 1;
        }
        else {
            my $OutputVersion = $Version;

            if ( $OutputVersion =~ m{ [0-9.] }xms ) {
                $OutputVersion = 'v' . $OutputVersion;
            }

            $Message .= "<green>ok</green> ($OutputVersion)\n$AdditionalText";
        }
    }
    else {
        my $Comment  = $Module->{Comment} ? ' - ' . $Module->{Comment} : '';
        my $Required = $Module->{Required};
        my $Color    = 'yellow';

        # OS Install Command
        my %InstallCommand = $Self->_GetInstallCommand($Module);

        # create example installation string for module
        my $InstallText = '';
        if ( IsHashRefWithData( \%InstallCommand ) ) {
            my $CMD = $InstallCommand{CMD};
            if ( $InstallCommand{SubCMD} ) {
                $CMD = sprintf $InstallCommand{CMD}, $InstallCommand{SubCMD};
            }

            $InstallText = "To install, you can use: '" . sprintf( $CMD, $InstallCommand{Package} ) . "'.";
        }

        if ($Required) {
            $Required = 'required';
            $Color    = 'red';
            $Error    = 1;
        }
        else {
            $Required = 'optional';
        }
        $Message .= "<$Color>Not installed!</$Color> ($Required$Comment)\n";
        $Message .= "$Indent  $InstallText \n";
    }

    if ( $Module->{Depends} ) {
        for my $ModuleSub ( @{ $Module->{Depends} } ) {
            my $DependencyCheck = $Self->_Check( $ModuleSub, $Depends + 1 );
            if ( !$DependencyCheck->{Success} ) {
                $Error = 1;
            }
            $Message .= $DependencyCheck->{Message};
        }
    }

    return {
        Success => $Error ? 0 : 1,
        Message => $Message,
    };
}

=head2 _PackageList()

Returns a hash with commands and a list of packages which can be installed with each command.

    my %PackagesByCommand = $PerlModulesObject->_PackageList(
        [
            {
                Module    => 'Archive::Tar',
                Required  => 1,
                Comment   => 'Required for compressed file generation (in perlcore).',
                InstTypes => {
                    emerge => 'perl-core/Archive-Tar',
                },
            },
            {
                Module    => 'Archive::Zip',
                Required  => 1,
                Comment   => 'Required for compressed file generation.',
                InstTypes => {
                    aptget => 'libarchive-zip-perl',
                    emerge => 'dev-perl/Archive-Zip',
                },
            },
            ...
        ],
    );

    Returns:

        my %PackagesByCommand = (
            'apt-get install -y' => {
                SubCMD => 'perl(%s)',
                Packages => [
                    'libarchive-zip-perl',
                    'libtimedate-perl',
                    'libdatetime-perl',
                ],
            },
            'cpan %s' => {
                Packages => [
                    'CryptX',
                    'Specio',
                    'Specio::Subs',
                ],
            },
        );

=cut

sub _PackageList {
    my ( $Self, $PackageList ) = @_;

    my %PackagesByCommand;

    MODULE:
    for my $Module ( @{$PackageList} ) {

        next MODULE if !$Module->{Required};

        my $Version = Kernel::System::Environment->ModuleVersionGet( Module => $Module->{Module} );

        next MODULE if $Version;

        my %InstallCommand = $Self->_GetInstallCommand($Module);

        if ( $Module->{Depends} ) {

            MODULESUB:
            for my $ModuleSub ( @{ $Module->{Depends} } ) {
                next MODULESUB if !$ModuleSub->{Required};

                my %InstallCommandSub = $Self->_GetInstallCommand($ModuleSub);
                next MODULESUB if !IsHashRefWithData( \%InstallCommandSub );

                push @{ $PackagesByCommand{ $InstallCommandSub{CMD} }->{Packages} }, $InstallCommandSub{Package};

                next MODULESUB if !$InstallCommandSub{SubCMD};
                $PackagesByCommand{ $InstallCommandSub{CMD} }->{SubCMD} = $InstallCommandSub{SubCMD};
            }
        }

        next MODULE if !IsHashRefWithData( \%InstallCommand );

        push @{ $PackagesByCommand{ $InstallCommand{CMD} }->{Packages} }, $InstallCommand{Package};

        next MODULE if !$InstallCommand{SubCMD};
        $PackagesByCommand{ $InstallCommand{CMD} }->{SubCMD} = $InstallCommand{SubCMD};
    }

    return %PackagesByCommand;
}

sub _GetInstallCommand {
    my ( $Self, $Module ) = @_;

    my $CMD;
    my $SubCMD;
    my $Package;

    # returns the installation type e.g. aptget
    my $InstType      = $Self->{DistToInstType}->{ $Self->{OSDist} };
    my $OutputInstall = 1;

    if ($InstType) {

        # gets the install command for installation type
        # e.g. apt-get install -y %s
        # default is the cpan install command
        # e.g. cpan %s
        $CMD    = $Self->{InstTypeToCMD}->{$InstType}->{CMD};
        $SubCMD = $Self->{InstTypeToCMD}->{$InstType}->{SubCMD};

        # gets the target package
        if (
            exists $Module->{InstTypes}->{$InstType}
            && !defined $Module->{InstTypes}->{$InstType}
            )
        {
            # if we a hash key for the installation type but a undefined value
            # then we prevent the output for the installation command
            $OutputInstall = 0;
        }
        elsif ( $Self->{InstTypeToCMD}->{$InstType}->{UseModule} ) {

            # default is the cpan module name
            $Package = $Module->{Module};
        }
        else {
            # if the package name is defined for the installation type
            # e.g. aptget then we use this as package name
            $Package = $Module->{InstTypes}->{$InstType};
        }
    }

    return if !$OutputInstall;

    if ( !$CMD || !$Package ) {
        $CMD     = $Self->{InstTypeToCMD}->{default}->{CMD};
        $SubCMD  = $Self->{InstTypeToCMD}->{default}->{SubCMD};
        $Package = $Module->{Module};
    }

    return (
        CMD     => $CMD,
        SubCMD  => $SubCMD,
        Package => $Package,
    );
}

1;
