#!/usr/bin/env perl
# --
# Copyright (C) 2001-2024 OTRS AG, https://otrs.com/
# --
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see https://www.gnu.org/licenses/gpl-3.0.txt.
# --

use strict;
use warnings;
use utf8;

use Getopt::Long;
use Term::ANSIColor;

## nofilter(TidyAll::Plugin::OTRS::Perl::BinScripts)

use File::Basename;
use FindBin qw($RealBin);
use lib dirname($RealBin);
use lib dirname($RealBin) . '/Kernel/cpan-lib';

use Kernel::System::CheckEnvironment;

my $Options;

#
# Please note that this script can be used in OTRS, or as a part of a standalone bundle
#   generated by Dev::Tools::GenerateCheckEnvironmentBundle.
#

Getopt::Long::GetOptions(
    'help',    \$Options->{Help},
    'all',     \$Options->{AllModules},
    'list',    \$Options->{PackageList},
    'no-ansi', \$Options->{NoANSI},
);

# Enable auto-flushing of STDOUT.
$| = 1;    ## no critic

if ( $Options->{Help} ) {
    Help();
    exit 1;
}

# Check if we are running in OTRS or from the standalone bundle.
if ( $Options->{AllModules} ) {
    if ( !-d dirname($RealBin) . "/Kernel/Config" ) {
        Print("<red>Checking bundled modules is not supported from the standalone bundle.</red>\n");
        exit 1;
    }
}

my $CheckEnvironmentObject = Kernel::System::CheckEnvironment->new();
my %CheckResult            = $CheckEnvironmentObject->Run(
    Options => $Options,
);

print "\n";

RESULT:
for my $Result ( @{ $CheckResult{Output} } ) {

    my $Color = '';
    if ( $Result->{State} eq 'Problem' || $Result->{State} eq 'Error' ) {
        $Color = 'red';
    }
    elsif ( uc( $Result->{State} ) eq 'OK' ) {
        $Color = 'green';
    }
    elsif ( $Result->{State} eq 'Info' ) {
        $Color = 'yellow';
    }

    next RESULT if !$Result->{Message};

    # If the message already contains color tags, use these and reset the state based color.
    if ( $Result->{Message} =~ m{ <(green|red|yellow)> }xms ) {
        $Color = '';
    }

    if ($Color) {
        Print("<$Color>$Result->{Message}</$Color>\n");
    }
    else {
        Print("$Result->{Message}\n");
    }
}

print "\n";

exit 0 if !$CheckResult{Error};

exit 1;

sub Help {
    my (%Param) = @_;

    Print(<<"EOF");

<green>otrs.CheckEnvironment.pl</green> - Script to check if environment and module dependencies are met for the OTRS application.
Copyright (C) 2001-2024 OTRS AG, https://otrs.com/

<yellow>Usage:</yellow>
 $0 [options]

<yellow>Options:</yellow>
 <green>[--help]</green>              - Display help for this script.
 <green>[--list]</green>              - Return an install command with all required packages.
 <green>[--all]</green>               - Return all required, optional and bundled packages of OTRS.
 <green>[--no-ansi]</green>           - Do not perform ANSI terminal output coloring.

EOF
    return 1;
}

=head2 Print()

This method will print the given text to STDOUT.

You can add color tags (C<< <green>...</green>, <yellow>...</yellow>, <red>...</red> >>) to your text, and they will be
replaced with the corresponding terminal escape sequences if the terminal supports it.

=cut

sub Print {
    my ($Text) = @_;

    print _ReplaceColorTags($Text);

    return;
}

=head2 _Color()

This will color the given text (see Term::ANSIColor::color()) if ANSI output is available and active, otherwise the text
stays unchanged.

    my $PossiblyColoredText = _Color('green', $Text);

=cut

sub _Color {
    my ( $Color, $Text ) = @_;

    return $Text if $Options->{NoANSI};

    return Term::ANSIColor::color($Color) . $Text . Term::ANSIColor::color('reset');
}

sub _ReplaceColorTags {
    my ($Text) = @_;

    $Text //= '';

    $Text =~ s{<(green|yellow|red)>(.*?)</\1>}{_Color($1, $2)}gsmxe;

    return $Text;
}
